# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Functions to help with hashing."""

from __future__ import annotations

import dataclasses
import hashlib
import typing as t
from collections.abc import Mapping

import aiofiles

if t.TYPE_CHECKING:
    from _typeshed import StrOrBytesPath


@dataclasses.dataclass(frozen=True)
class _AlgorithmData:
    name: str
    algorithm: str
    kwargs: dict[str, t.Any]


_PREFERRED_HASHES: tuple[_AlgorithmData, ...] = (
    # https://pypi.org/help/#verify-hashes, https://github.com/pypi/warehouse/issues/9628
    _AlgorithmData(name="sha256", algorithm="sha256", kwargs={}),
    _AlgorithmData(name="blake2b_256", algorithm="blake2b", kwargs={"digest_size": 32}),
)


async def verify_hash(
    filename: StrOrBytesPath,
    hash_digest: str,
    *,
    algorithm: str = "sha256",
    algorithm_kwargs: dict[str, t.Any] | None = None,
    chunksize: int,
) -> bool:
    """
    Verify whether a file has a given sha256sum.

    :arg filename: The file to verify the sha256sum of.
    :arg hash_digest: The hash that is expected.
    :kwarg algorithm: The hash algorithm to use.  This must be present in hashlib on this
        system.  The default is 'sha256'
    :kwarg algorithm_kwargs: Parameters to provide to the hash algorithm's constructor.
    :returns: True if the hash matches, otherwise False.
    """
    hasher = getattr(hashlib, algorithm)(**(algorithm_kwargs or {}))
    async with aiofiles.open(filename, "rb") as f:
        while chunk := await f.read(chunksize):
            hasher.update(chunk)
    if hasher.hexdigest() != hash_digest:
        return False

    return True


async def verify_a_hash(
    filename: StrOrBytesPath,
    hash_digests: Mapping[str, str],
    *,
    chunksize: int,
) -> bool:
    """
    Verify whether a file has a given hash, given a set of digests with different algorithms.
    Will only test trustworthy hashes and return ``False`` if none matches.

    :arg filename: The file to verify the hash of.
    :arg hash_digest: A mapping of hash types to digests.
    :returns: True if the hash matches, otherwise False.
    """
    for algorithm_data in _PREFERRED_HASHES:
        if algorithm_data.name in hash_digests:
            return await verify_hash(
                filename,
                hash_digests[algorithm_data.name],
                algorithm=algorithm_data.algorithm,
                algorithm_kwargs=algorithm_data.kwargs,
                chunksize=chunksize,
            )
    return False



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2021, Ansible Project

"""
YAML handling.
"""

from __future__ import annotations

import typing as t

import yaml

_SafeLoader: t.Any
_SafeDumper: t.Any
try:
    # use C version if possible for speedup
    from yaml import CSafeDumper as _SafeDumper
    from yaml import CSafeLoader as _SafeLoader
except ImportError:  # pragma: no cover
    from yaml import SafeDumper as _SafeDumper
    from yaml import SafeLoader as _SafeLoader

if t.TYPE_CHECKING:
    from _typeshed import StrOrBytesPath, SupportsWrite


class _IndentedDumper(yaml.SafeDumper):
    """
    Extend YAML dumper to increase indent of list items.
    """

    def increase_indent(self, flow=False, indentless=False):
        return super().increase_indent(flow, False)


def load_yaml_bytes(data: bytes) -> t.Any:
    """
    Load and parse YAML from given bytes.
    """
    return yaml.load(data, Loader=_SafeLoader)


def load_yaml_file(path: StrOrBytesPath) -> t.Any:
    """
    Load and parse YAML file ``path``.
    """
    with open(path, "rb") as stream:
        return yaml.load(stream, Loader=_SafeLoader)


def store_yaml_file(
    path: StrOrBytesPath, content: t.Any, *, nice: bool = False, sort_keys: bool = True
) -> None:
    """
    Store ``content`` as YAML file under ``path``.
    """
    with open(path, "wb") as stream:
        store_yaml_stream(stream, content, nice=nice, sort_keys=sort_keys)


def store_yaml_stream(
    stream: SupportsWrite, content: t.Any, *, nice: bool = False, sort_keys: bool = True
) -> None:
    """
    Dump ``content`` as YAML to an IO ``stream``.
    """
    dumper = _IndentedDumper if nice else _SafeDumper
    dumper.ignore_aliases = lambda *args: True  # type: ignore[method-assign]
    yaml.dump(
        content,
        stream,
        default_flow_style=False,
        encoding="utf-8",
        Dumper=dumper,
        explicit_start=nice,
        sort_keys=sort_keys,
    )



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project

from __future__ import annotations

import pytest

from antsibull_fileutils.hashing import verify_a_hash, verify_hash

HASH_TESTS = [
    (
        b"",
        "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
        "sha256",
        {},
        True,
    ),
    (
        b"",
        "01ba4719c80b6fe911b091a7c05124b64eeece964e09c058ef8f9805daca546b",
        "sha256",
        {},
        False,
    ),
]


@pytest.mark.parametrize(
    "content, hash, algorithm, algorithm_kwargs, expected_match", HASH_TESTS
)
@pytest.mark.asyncio
async def test_verify_hash(
    content: bytes,
    hash: bytes,
    algorithm: str,
    algorithm_kwargs: dict | None,
    expected_match: bool,
    tmp_path,
):
    filename = tmp_path / "file"
    with open(filename, "wb") as f:
        f.write(content)
    result = await verify_hash(
        filename,
        hash,
        algorithm=algorithm,
        algorithm_kwargs=algorithm_kwargs,
        chunksize=65536,
    )
    assert result is expected_match


HASH_DICT_TESTS = [
    (
        b"foo",
        {
            "sha256": "2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae",
        },
        True,
    ),
    (
        b"bar",
        {
            "sha256": "2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae",
        },
        False,
    ),
    (
        b"foo",
        {
            "blake2b_256": "b8fe9f7f6255a6fa08f668ab632a8d081ad87983c77cd274e48ce450f0b349fd",
        },
        True,
    ),
    (
        b"bar",
        {
            "blake2b_256": "b8fe9f7f6255a6fa08f668ab632a8d081ad87983c77cd274e48ce450f0b349fd",
        },
        False,
    ),
    (
        b"",
        {},
        False,
    ),
    (
        b"",
        {
            "foo": "bar",
        },
        False,
    ),
]


@pytest.mark.parametrize("content, hash_digests, expected_match", HASH_DICT_TESTS)
@pytest.mark.asyncio
async def test_verify_a_hash(
    content: bytes, hash_digests: dict[str, str], expected_match: bool, tmp_path
):
    filename = tmp_path / "file"
    with open(filename, "wb") as f:
        f.write(content)
    result = await verify_a_hash(filename, hash_digests, chunksize=65536)
    assert result is expected_match



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
Test vcs module.
"""

from __future__ import annotations

import subprocess
from unittest import mock

import pytest

from antsibull_fileutils.vcs import detect_vcs, list_git_files

from .utils import collect_log


def test_detect_vcs():
    path = "/path/to/dir"
    git_bin_path = "/path/to/git"
    git_command = [git_bin_path, "-C", path, "rev-parse", "--is-inside-work-tree"]

    with mock.patch(
        "subprocess.check_output",
        return_value="true\n",
    ) as m:
        kwargs, debug, info = collect_log()
        assert detect_vcs(path, git_bin_path=git_bin_path, **kwargs) == "git"
        m.assert_called_with(
            git_command, text=True, encoding="utf-8", stderr=subprocess.PIPE
        )
        assert debug == [
            ("Trying to determine whether {!r} is a Git repository", ("/path/to/dir",)),
            ("Git output: {}", ("true",)),
        ]
        assert info == [
            ("Identified {!r} as a Git repository", ("/path/to/dir",)),
        ]

    with mock.patch(
        "subprocess.check_output",
        return_value="true\n",
    ) as m:
        assert detect_vcs(path, git_bin_path=git_bin_path) == "git"
        m.assert_called_with(
            git_command, text=True, encoding="utf-8", stderr=subprocess.PIPE
        )

    with mock.patch(
        "subprocess.check_output",
        return_value="foobar\n".encode("utf-8"),
    ) as m:
        assert detect_vcs(path, git_bin_path=git_bin_path) == "none"
        m.assert_called_with(
            git_command, text=True, encoding="utf-8", stderr=subprocess.PIPE
        )

    with mock.patch(
        "subprocess.check_output",
        side_effect=subprocess.CalledProcessError(128, path),
    ) as m:
        assert detect_vcs(path, git_bin_path=git_bin_path) == "none"
        m.assert_called_with(
            git_command, text=True, encoding="utf-8", stderr=subprocess.PIPE
        )

    with mock.patch(
        "subprocess.check_output",
        side_effect=FileNotFoundError(),
    ) as m:
        assert detect_vcs(path, git_bin_path=git_bin_path) == "none"
        m.assert_called_with(
            git_command, text=True, encoding="utf-8", stderr=subprocess.PIPE
        )


TEST_LIST_GIT_FILES = [
    (b"", [], False),
    (b"", [], True),
    (b"foo\nbar", [b"foo\nbar"], False),
    (b"foo\x00", [b"foo"], True),
    (b"foo\x00bar", [b"foo", b"bar"], True),
    (
        b"link\x00foobar\x00dir/binary_file",
        [b"link", b"foobar", b"dir/binary_file"],
        False,
    ),
]


@pytest.mark.parametrize("stdout, expected, with_logging", TEST_LIST_GIT_FILES)
def test_list_git_files(stdout: bytes, expected: list[str], with_logging: bool):
    path = "/path/to/dir"
    git_bin_path = "/path/to/git"
    git_command = [
        git_bin_path,
        "ls-files",
        "-z",
        "--cached",
        "--others",
        "--exclude-standard",
        "--deduplicate",
    ]

    with mock.patch(
        "subprocess.check_output",
        return_value=stdout,
    ) as m:
        kwargs, debug, info = collect_log(with_debug=with_logging, with_info=False)
        assert list_git_files(path, git_bin_path=git_bin_path, **kwargs) == expected
        m.assert_called_with(git_command, cwd=path, stderr=subprocess.PIPE)
        if with_logging:
            assert debug == [("Identifying files not ignored by Git in {!r}", (path,))]


def test_list_git_files_fail():
    path = "/path/to/dir"
    git_bin_path = "/path/to/git"
    git_command = [
        git_bin_path,
        "ls-files",
        "-z",
        "--cached",
        "--others",
        "--exclude-standard",
        "--deduplicate",
    ]

    with mock.patch(
        "subprocess.check_output",
        side_effect=subprocess.CalledProcessError(128, path),
    ) as m:
        with pytest.raises(ValueError, match="^Error while running git$") as exc:
            list_git_files(path, git_bin_path=git_bin_path)

    with mock.patch(
        "subprocess.check_output",
        side_effect=FileNotFoundError(),
    ) as m:
        with pytest.raises(ValueError, match="^Cannot find git executable$") as exc:
            list_git_files(path, git_bin_path=git_bin_path)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
Directory and collection copying helpers.
"""

from __future__ import annotations

import os
import shutil
import typing as t

from antsibull_fileutils.tempfile import ansible_mkdtemp
from antsibull_fileutils.vcs import list_git_files

if t.TYPE_CHECKING:
    from _typeshed import StrPath


class CopierError(Exception):
    pass


def _is_internal(directory: str, link: str) -> bool:
    dest = os.path.join(directory, link)
    if os.path.isabs(dest):
        return False

    if os.path.splitdrive(dest)[0]:
        return False

    normpath = os.path.normpath(dest)
    return not (normpath == ".." or normpath.startswith(".." + os.sep))


class _TreeCopier:
    def __init__(
        self,
        source: StrPath,
        dest: StrPath,
        *,
        keep_inside_symlinks: bool = True,
        keep_outside_symlinks: bool = False,
        normalize_links: bool = True,
        log_debug: t.Callable[[str], None] | None = None,
    ):
        """
        Initialize copy helper
        """
        self._log_debug = log_debug
        self.created_directories: set[str] = {".", ""}
        self.source = source
        self.dest = dest
        self.keep_inside_symlinks = keep_inside_symlinks
        self.keep_outside_symlinks = keep_outside_symlinks
        self.never_keep = not (keep_inside_symlinks or keep_outside_symlinks)
        self.normalize_links = normalize_links
        os.mkdir(self.dest, mode=0o700)

    def _do_log_debug(self, msg: str, *args: t.Any) -> None:
        if self._log_debug:
            self._log_debug(msg, *args)

    def _copy_link(self, directory: str, full_source: str, full_dest: str) -> None:
        link = os.readlink(full_source)
        if self.normalize_links:
            full_directory = os.path.join(self.source, directory)
            link = os.path.relpath(os.path.join(full_directory, link), full_directory)

        internal = False if self.never_keep else _is_internal(directory, link)
        keep = self.keep_inside_symlinks if internal else self.keep_outside_symlinks
        if keep:
            self._do_log_debug("Copying symlink {!r} to {!r}", full_source, full_dest)
            os.symlink(link, full_dest)
            shutil.copystat(full_source, full_dest, follow_symlinks=False)
            return

        real_source = os.path.realpath(full_source)
        if os.path.isdir(real_source):
            self._do_log_debug(
                "Copying symlinked directory tree {!r} to {!r}", full_source, full_dest
            )
            shutil.copytree(real_source, full_dest, symlinks=False)
            return

        self._do_log_debug(
            "Copying symlinked file {!r} to {!r}", full_source, full_dest
        )
        shutil.copy2(real_source, full_dest)

    def _create_dir(self, directory: str) -> None:
        if directory not in self.created_directories:
            src_dir = os.path.join(self.source, directory)
            dest_dir = os.path.join(self.dest, directory)
            self._do_log_debug("Copying directory {!r} to {!r}", src_dir, dest_dir)
            os.makedirs(dest_dir, mode=0o700, exist_ok=True)
            shutil.copystat(src_dir, dest_dir, follow_symlinks=False)
            self.created_directories.add(directory)

    def _copy_file(self, directory: str, relative_path: str) -> None:
        self._create_dir(directory)

        full_source = os.path.join(self.source, relative_path)
        full_dest = os.path.join(self.dest, relative_path)
        if os.path.islink(full_source):
            self._copy_link(directory, full_source, full_dest)
        else:
            self._do_log_debug("Copying file {!r} to {!r}", full_source, full_dest)
            shutil.copy2(full_source, full_dest)

    def copy_file(
        self, relative_path: str, *, ignore_non_existing: bool = False
    ) -> None:
        if ignore_non_existing and not os.path.lexists(
            os.path.join(self.source, relative_path)
        ):
            return
        directory, _ = os.path.split(relative_path)
        self._copy_file(directory, relative_path)

    def walk(self, exclude_root: list[str] | None = None):
        exclude_root_set: set[str] = set(exclude_root or [])
        for root, dirs, files in os.walk(self.source, followlinks=False, topdown=True):
            directory = os.path.relpath(root, self.source)
            if directory == ".":
                directory = ""
            for file in files:
                if file in exclude_root_set:
                    continue
                relative_path = os.path.join(directory, file)
                self._copy_file(directory, relative_path)
            for path in exclude_root_set:
                try:
                    dirs.remove(path)
                except ValueError:
                    pass
            for a_dir in dirs:
                self._create_dir(os.path.join(directory, a_dir))
            exclude_root_set.clear()


class Copier:
    """
    Allows to copy directories.
    """

    def __init__(
        self,
        *,
        normalize_links: bool = True,
        log_debug: t.Callable[[str], None] | None = None,
    ):
        self.normalize_links = normalize_links
        self._log_debug = log_debug

    def _do_log_debug(self, msg: str, *args: t.Any) -> None:
        if self._log_debug:
            self._log_debug(msg, *args)

    def copy(
        self,
        from_path: StrPath,
        to_path: StrPath,
        *,
        exclude_root: list[str] | None = None,
    ) -> None:
        """
        Copy a directory ``from_path`` to a destination ``to_path``.

        ``to_path`` must not exist, but its parent directory must exist.
        """
        self._do_log_debug(
            "Copying complete directory from {!r} to {!r}", from_path, to_path
        )
        _TreeCopier(
            from_path,
            to_path,
            normalize_links=self.normalize_links,
            log_debug=self._log_debug,
        ).walk(exclude_root=exclude_root)


class GitCopier(Copier):
    """
    Allows to copy directories that are part of a Git repository.
    """

    def __init__(
        self,
        *,
        normalize_links: bool = True,
        git_bin_path: StrPath = "git",
        log_debug: t.Callable[[str], None] | None = None,
        copy_repo_structure: bool = False,
    ):
        super().__init__(normalize_links=normalize_links, log_debug=log_debug)
        self.git_bin_path = git_bin_path
        self.copy_repo_structure = copy_repo_structure

    def copy(
        self,
        from_path: StrPath,
        to_path: StrPath,
        *,
        exclude_root: list[str] | None = None,
    ) -> None:
        self._do_log_debug("Identifying files not ignored by Git in {!r}", from_path)
        try:
            files = list_git_files(
                from_path, git_bin_path=self.git_bin_path, log_debug=self._log_debug
            )
        except ValueError as exc:
            raise CopierError(
                f"Error while listing files not ignored by Git in {from_path}: {exc}"
            ) from exc

        self._do_log_debug(
            "Copying {} file(s) from {!r} to {!r}", len(files), from_path, to_path
        )
        tc = _TreeCopier(
            from_path,
            to_path,
            normalize_links=self.normalize_links,
            log_debug=self._log_debug,
        )
        exclude_root_set = set(exclude_root or [])
        exclude_root_prefixes = [f"{path}/" for path in exclude_root or []]
        for file in files:
            # Decode filename and check whether the file still exists
            # (deleted files are part of the output)
            file_decoded = file.decode("utf-8")
            if file_decoded in exclude_root_set:
                continue
            if any(file_decoded.startswith(prefix) for prefix in exclude_root_prefixes):
                continue
            tc.copy_file(file_decoded, ignore_non_existing=True)
        # Copy .git directory as well if requested
        if self.copy_repo_structure and os.path.isdir(os.path.join(from_path, ".git")):
            _TreeCopier(
                os.path.join(from_path, ".git"),
                os.path.join(to_path, ".git"),
                normalize_links=self.normalize_links,
                log_debug=self._log_debug,
            ).walk()


class CollectionCopier:
    """
    Creates a copy of a collection to a place where ``--playbook-dir`` can be used
    to prefer this copy of the collection over any installed ones.
    """

    def __init__(
        self,
        *,
        source_directory: StrPath,
        namespace: str,
        name: str,
        copier: Copier,
        log_debug: t.Callable[[str], None] | None = None,
    ):
        self.source_directory = source_directory
        self.namespace = namespace
        self.name = name
        self.copier = copier
        self._log_debug = log_debug

        self.dir = os.path.realpath(ansible_mkdtemp(prefix="antsibull-fileutils"))

    def _do_log_debug(self, msg: str, *args: t.Any) -> None:
        if self._log_debug:
            self._log_debug(msg, *args)

    def __enter__(self) -> tuple[str, str]:
        try:
            collection_container_dir = os.path.join(
                self.dir, "collections", "ansible_collections", self.namespace
            )
            os.makedirs(collection_container_dir)

            collection_dir = os.path.join(collection_container_dir, self.name)
            self._do_log_debug("Temporary collection directory: {!r}", collection_dir)

            self.copier.copy(self.source_directory, collection_dir)

            self._do_log_debug("Temporary collection directory has been populated")
            return (
                self.dir,
                collection_dir,
            )
        except Exception:
            shutil.rmtree(self.dir, ignore_errors=True)
            raise

    def __exit__(self, type_, value, traceback_):
        shutil.rmtree(self.dir, ignore_errors=True)



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project

from __future__ import annotations

from io import BytesIO

import pytest

from antsibull_fileutils.yaml import (
    load_yaml_bytes,
    load_yaml_file,
    store_yaml_file,
    store_yaml_stream,
)

LOAD_YAML_DATA = [
    (
        """
foo: bar
""",
        {
            "foo": "bar",
        },
    ),
]


@pytest.mark.parametrize(
    "content, expected",
    LOAD_YAML_DATA,
)
def test_load_yaml(content, expected, tmp_path):
    assert load_yaml_bytes(content.encode("utf-8")) == expected

    file = tmp_path / "test.yaml"
    with file.open("w", encoding="utf-8") as f:
        f.write(content)
    assert load_yaml_file(file) == expected
    assert load_yaml_file(str(file)) == expected


STORE_YAML_DATA = [
    (
        {"foo": "bar", "baz": ["bam", "foobar", "aaa"]},
        {},
        """baz:
- bam
- foobar
- aaa
foo: bar
""",
    ),
    (
        {"foo": "bar", "baz": ["bam", "foobar", "aaa"]},
        {
            "nice": True,
        },
        """---
baz:
  - bam
  - foobar
  - aaa
foo: bar
""",
    ),
    (
        {"foo": "bar", "baz": ["bam", "foobar", "aaa"]},
        {
            "sort_keys": False,
        },
        """foo: bar
baz:
- bam
- foobar
- aaa
""",
    ),
]


@pytest.mark.parametrize(
    "content, kwargs, expected",
    STORE_YAML_DATA,
)
def test_store_yaml(content, kwargs, expected, tmp_path):
    file = tmp_path / "test.yaml"

    store_yaml_file(file, content, **kwargs)
    with file.open("r", encoding="utf-8") as f:
        assert f.read() == expected

    store_yaml_file(str(file), content, **kwargs)
    with file.open("r", encoding="utf-8") as f:
        assert f.read() == expected

    stream = BytesIO()
    store_yaml_stream(stream, content, **kwargs)
    assert stream.getvalue().decode("utf-8") == expected



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2021, Ansible Project
"""I/O helper functions."""

from __future__ import annotations

import os
import os.path
import typing as t

import aiofiles

if t.TYPE_CHECKING:
    from _typeshed import StrOrBytesPath


async def copy_file(
    source_path: StrOrBytesPath,
    dest_path: StrOrBytesPath,
    *,
    check_content: bool = True,
    file_check_content: int = 0,
    chunksize: int,
) -> bool:
    """
    Copy content from one file to another.

    :arg source_path: Source path. Must be a file.
    :arg dest_path: Destination path.
    :kwarg check_content: If ``True`` (default) and ``file_check_content > 0`` and the
        destination file exists, first check whether source and destination are potentially equal
        before actually copying,
    :return: ``True`` if the file was actually copied.
    """
    if check_content and file_check_content > 0:
        # Check whether the destination file exists and has the same content as the source file,
        # in which case we won't overwrite the destination file
        try:
            stat_d = os.stat(dest_path)
            if stat_d.st_size <= file_check_content:
                stat_s = os.stat(source_path)
                if stat_d.st_size == stat_s.st_size:
                    # Read both files and compare
                    async with aiofiles.open(source_path, "rb") as f_in:
                        content_to_copy = await f_in.read()
                    async with aiofiles.open(dest_path, "rb") as f_in:
                        existing_content = await f_in.read()
                    if content_to_copy == existing_content:
                        return False
                    # Since we already read the contents of the file to copy, simply write it to
                    # the destination instead of reading it again
                    async with aiofiles.open(dest_path, "wb") as f_out:
                        await f_out.write(content_to_copy)
                    return True
        except FileNotFoundError:
            # Destination (or source) file does not exist
            pass

    async with aiofiles.open(source_path, "rb") as f_in:
        async with aiofiles.open(dest_path, "wb") as f_out:
            while chunk := await f_in.read(chunksize):
                await f_out.write(chunk)
    return True


async def write_file(
    filename: StrOrBytesPath,
    content: str,
    *,
    file_check_content: int = 0,
    encoding: str = "utf-8",
) -> bool:
    """
    Write encoded content to file.

    :arg filename: The filename to write to.
    :arg content: The content to write to the file.
    :kwarg file_check_content: If > 0 and the file exists and its size in bytes does not exceed this
        value, will read the file and compare it to the encoded content before overwriting.
    :return: ``True`` if the file was actually written.
    """

    content_bytes = content.encode(encoding)

    if file_check_content > 0 and len(content_bytes) <= file_check_content:
        # Check whether the destination file exists and has the same content as the one we want to
        # write, in which case we won't overwrite the file
        try:
            stat = os.stat(filename)
            if stat.st_size == len(content_bytes):
                # Read file and compare
                async with aiofiles.open(filename, "rb") as f:
                    existing_content = await f.read()
                if existing_content == content_bytes:
                    return False
        except FileNotFoundError:
            # Destination file does not exist
            pass

    async with aiofiles.open(filename, "wb") as f:
        await f.write(content_bytes)
    return True


async def read_file(filename: StrOrBytesPath, *, encoding: str = "utf-8") -> str:
    """
    Read the file and decode its contents with the given encoding.

    :arg filename: The filename to read from.
    :kwarg encoding: The encoding to use.
    """
    async with aiofiles.open(filename, "r", encoding=encoding) as f:
        content = await f.read()

    return content



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project

from __future__ import annotations

from io import BytesIO

import pytest

from antsibull_fileutils.io import copy_file, read_file, write_file


@pytest.mark.asyncio
async def test_copy_file(tmp_path):
    content = "foo\x00bar\x00baz\x00çäø☺".encode("utf-8")
    content_len = len(content)

    alt_content = "boo\x00bar\x00baz\x00çäø☺".encode("utf-8")
    alt_content_len = "foo\x00bar".encode("utf-8")

    src_path = tmp_path / "file-src"
    src_path.write_bytes(content)

    dst_path = tmp_path / "file-dst"
    assert await copy_file(src_path, dst_path, chunksize=4) is True
    assert dst_path.read_bytes() == content

    assert (
        await copy_file(
            src_path, dst_path, chunksize=4, file_check_content=content_len - 1
        )
        is True
    )
    assert dst_path.read_bytes() == content

    assert (
        await copy_file(src_path, dst_path, chunksize=4, file_check_content=content_len)
        is False
    )
    assert dst_path.read_bytes() == content

    src_path.write_bytes(alt_content)

    assert (
        await copy_file(src_path, dst_path, chunksize=4, file_check_content=content_len)
        is True
    )
    assert dst_path.read_bytes() == alt_content

    src_path.write_bytes(alt_content_len)

    assert (
        await copy_file(src_path, dst_path, chunksize=4, file_check_content=content_len)
        is True
    )
    assert dst_path.read_bytes() == alt_content_len

    dst_path = tmp_path / "file-dst-2"
    assert (
        await copy_file(src_path, dst_path, chunksize=4, file_check_content=content_len)
        is True
    )
    assert dst_path.read_bytes() == alt_content_len


@pytest.mark.asyncio
async def test_read_file(tmp_path):
    content = "foo\x00bar\x00baz\x00çäø☺"
    filename = tmp_path / "file"
    filename.write_text(content, encoding="utf-8")

    assert await read_file(filename, encoding="utf-8") == content
    assert await read_file(str(filename), encoding="utf-8") == content
    assert await read_file(str(filename).encode("utf-8"), encoding="utf-8") == content
    assert await read_file(str(filename).encode("utf-8"), encoding="utf-8") == content

    filename.write_text(content, encoding="utf-16")

    assert await read_file(filename, encoding="utf-16") == content


@pytest.mark.asyncio
async def test_write_file(tmp_path):
    content = "foo\x00bar\x00baz\x00çäø☺"
    encoded_content_len = len(content.encode("utf-8"))

    alt_content = "boo\x00bar\x00baz\x00çäø☺"
    alt_content_len = "foo\x00bar"

    filename = tmp_path / "file"
    alt_filename = tmp_path / "file2"

    assert await write_file(filename, content, encoding="utf-8") is True
    assert filename.read_text(encoding="utf-8") == content

    assert await write_file(filename, content, encoding="utf-8") is True
    assert filename.read_text(encoding="utf-8") == content

    assert (
        await write_file(
            filename,
            content,
            encoding="utf-8",
            file_check_content=encoded_content_len - 1,
        )
        is True
    )
    assert filename.read_text(encoding="utf-8") == content

    assert (
        await write_file(
            filename, content, encoding="utf-8", file_check_content=encoded_content_len
        )
        is False
    )
    assert filename.read_text(encoding="utf-8") == content

    assert (
        await write_file(
            filename,
            content,
            encoding="utf-8",
            file_check_content=encoded_content_len + 1,
        )
        is False
    )
    assert filename.read_text(encoding="utf-8") == content

    assert (
        await write_file(
            filename,
            alt_content,
            encoding="utf-8",
            file_check_content=encoded_content_len,
        )
        is True
    )
    assert filename.read_text(encoding="utf-8") == alt_content

    assert (
        await write_file(
            alt_filename,
            alt_content,
            encoding="utf-8",
            file_check_content=encoded_content_len,
        )
        is True
    )
    assert alt_filename.read_text(encoding="utf-8") == alt_content

    assert (
        await write_file(
            alt_filename,
            alt_content_len,
            encoding="utf-8",
            file_check_content=encoded_content_len,
        )
        is True
    )
    assert alt_filename.read_text(encoding="utf-8") == alt_content_len



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
Test utils module.
"""

from __future__ import annotations

import os
import pathlib
import re
from unittest import mock

import pytest

from antsibull_fileutils.copier import (
    CollectionCopier,
    Copier,
    CopierError,
    GitCopier,
    _is_internal,
)

from .utils import collect_log


@pytest.mark.parametrize(
    "directory, link, expected",
    [
        ("", "foo", True),
        ("", "foo/bar", True),
        ("", "foo/../bar", True),
        ("", "foo/../../bar", False),
        ("", "..", False),
        ("", "../foo/bar", False),
        ("foo", ".", True),
        ("foo", "bar", True),
        ("foo", "../bar", True),
        ("foo", "../../bar", False),
        ("foo", "..", True),
    ],
)
def test__is_internal(directory, link, expected):
    assert _is_internal(directory, link) == expected


def assert_same(
    a: pathlib.Path, b: pathlib.Path, *, changed_link: str | None = None
) -> None:
    if a.is_symlink():
        assert b.is_symlink()
        assert (changed_link or a.readlink()) == b.readlink()
        return
    assert not b.is_symlink()
    if a.is_file():
        assert b.is_file()
        assert a.read_bytes() == b.read_bytes()
        return
    if a.is_dir():
        assert b.is_dir()
        return


def assert_same_recursively(a: pathlib.Path, b: pathlib.Path) -> None:
    assert a.is_dir()
    assert b.is_dir()
    dest = {f.name: f for f in b.iterdir()}
    for f in a.iterdir():
        f_other = dest.pop(f.name)
        assert_same(f, f_other)
        if f.is_dir():
            assert_same_recursively(f, f_other)
    assert dest == {}


def test_copier(tmp_path_factory):
    directory: pathlib.Path = tmp_path_factory.mktemp("copier")

    src_dir = directory / "src"
    src_dir.mkdir()
    (src_dir / "empty").touch()
    (src_dir / "link").symlink_to("empty")
    (src_dir / "dir").mkdir()
    (src_dir / "file").write_text("content", encoding="utf-8")
    (src_dir / "dir" / "binary_file").write_bytes(b"\x00\x01\x02")
    (src_dir / "dir" / "another_file").write_text("more", encoding="utf-8")

    dest_dir = directory / "dest1"
    copier = Copier()
    copier.copy(str(src_dir), str(dest_dir))
    assert_same_recursively(src_dir, dest_dir)

    def src_dst(*appendix):
        s = src_dir
        d = dest_dir
        for a in appendix:
            s = s / a
            d = d / a
        return str(s), str(d)

    dest_dir = directory / "dest2"
    kwargs, debug, info = collect_log(with_info=False)
    copier = Copier(**kwargs)
    copier.copy(
        str(src_dir),
        str(dest_dir),
    )
    assert_same_recursively(src_dir, dest_dir)
    assert sorted(debug) == [
        (
            "Copying complete directory from {!r} to {!r}",
            (str(src_dir), str(dest_dir)),
        ),
        (
            "Copying directory {!r} to {!r}",
            src_dst("dir"),
        ),
        (
            "Copying file {!r} to {!r}",
            src_dst("dir", "another_file"),
        ),
        (
            "Copying file {!r} to {!r}",
            src_dst("dir", "binary_file"),
        ),
        (
            "Copying file {!r} to {!r}",
            src_dst("empty"),
        ),
        (
            "Copying file {!r} to {!r}",
            src_dst("file"),
        ),
        (
            "Copying symlink {!r} to {!r}",
            src_dst("link"),
        ),
    ]

    dest_dir = directory / "dest3"
    kwargs, debug, info = collect_log(with_info=False)
    copier = Copier(**kwargs)
    copier.copy(src_dir, dest_dir)
    assert_same_recursively(src_dir, dest_dir)
    assert sorted(debug) == [
        ("Copying complete directory from {!r} to {!r}", (src_dir, dest_dir)),
        (
            "Copying directory {!r} to {!r}",
            src_dst("dir"),
        ),
        (
            "Copying file {!r} to {!r}",
            src_dst("dir", "another_file"),
        ),
        (
            "Copying file {!r} to {!r}",
            src_dst("dir", "binary_file"),
        ),
        (
            "Copying file {!r} to {!r}",
            src_dst("empty"),
        ),
        (
            "Copying file {!r} to {!r}",
            src_dst("file"),
        ),
        (
            "Copying symlink {!r} to {!r}",
            src_dst("link"),
        ),
    ]

    dest_dir = directory / "dest4"
    kwargs, debug, info = collect_log(with_info=False)
    copier = Copier(**kwargs)
    copier.copy(src_dir, dest_dir, exclude_root=["does-not-exist", "dir", "file"])
    assert sorted(debug) == [
        ("Copying complete directory from {!r} to {!r}", (src_dir, dest_dir)),
        (
            "Copying file {!r} to {!r}",
            src_dst("empty"),
        ),
        (
            "Copying symlink {!r} to {!r}",
            src_dst("link"),
        ),
    ]


def test_git_copier(tmp_path_factory):
    directory: pathlib.Path = tmp_path_factory.mktemp("git-copier")

    (directory / "other").write_bytes(b"other")
    (directory / "other_dir").mkdir()
    (directory / "other_dir" / "file").write_bytes(b"foo")

    src_dir = directory / "src"
    src_dir.mkdir()
    (src_dir / "empty").touch()
    (src_dir / "link").symlink_to("empty")
    (src_dir / "link_dir").symlink_to("dir")
    (src_dir / "trick_link").symlink_to("../src/empty")
    (src_dir / "out_link").symlink_to("../other")
    (src_dir / "out_link_dir").symlink_to("../other_dir")
    (src_dir / "abs_link").symlink_to(str((src_dir / "link").resolve()))
    (src_dir / "dead_link").symlink_to("does-not-exist")
    (src_dir / "out_dead_link").symlink_to("../does-not-exist")
    (src_dir / "dir").mkdir()
    (src_dir / "file").write_text("content", encoding="utf-8")
    (src_dir / "dir" / "binary_file").write_bytes(b"\x00\x01\x02")
    (src_dir / "dir" / "another_file").write_text("more", encoding="utf-8")

    dest_dir = directory / "dest1"
    with mock.patch(
        "antsibull_fileutils.copier.list_git_files",
        return_value=[b"file"],
    ) as m:
        copier = GitCopier(git_bin_path="/path/to/git")
        copier.copy(str(src_dir), str(dest_dir))
        m.assert_called_with(str(src_dir), git_bin_path="/path/to/git", log_debug=None)
        assert dest_dir.is_dir()
        assert {p.name for p in dest_dir.iterdir()} == {"file"}
        assert_same(src_dir / "file", dest_dir / "file")

    def src_dst(*appendix):
        s = src_dir
        d = dest_dir
        for a in appendix:
            s = s / a
            d = d / a
        return str(s), str(d)

    dest_dir = directory / "dest2"
    with mock.patch(
        "antsibull_fileutils.copier.list_git_files",
        return_value=[b"link", b"foobar", b"dir/binary_file", b"dir/another_file"],
    ) as m:
        kwargs, debug, info = collect_log(with_info=False)
        copier = GitCopier(git_bin_path="/path/to/git", **kwargs)
        copier.copy(src_dir, dest_dir)
        m.assert_called_with(src_dir, git_bin_path="/path/to/git", **kwargs)

        assert debug == [
            ("Identifying files not ignored by Git in {!r}", (src_dir,)),
            ("Copying {} file(s) from {!r} to {!r}", (4, src_dir, dest_dir)),
            ("Copying symlink {!r} to {!r}", src_dst("link")),
            ("Copying directory {!r} to {!r}", src_dst("dir")),
            ("Copying file {!r} to {!r}", src_dst("dir", "binary_file")),
            ("Copying file {!r} to {!r}", src_dst("dir", "another_file")),
        ]
        assert dest_dir.is_dir()
        assert {p.name for p in dest_dir.iterdir()} == {"link", "dir"}
        assert_same(src_dir / "link", dest_dir / "link")
        assert_same(src_dir / "dir", dest_dir / "dir")
        assert {p.name for p in (dest_dir / "dir").iterdir()} == {
            "another_file",
            "binary_file",
        }
        assert_same(src_dir / "dir" / "another_file", dest_dir / "dir" / "another_file")
        assert_same(src_dir / "dir" / "binary_file", dest_dir / "dir" / "binary_file")

    dest_dir = directory / "dest3"
    with mock.patch(
        "antsibull_fileutils.copier.list_git_files",
        return_value=[
            b"link_dir",
            b"trick_link",
            b"out_link",
            b"out_link_dir",
            b"abs_link",
            b"dead_link",
        ],
    ) as m:
        copier = GitCopier(git_bin_path="/path/to/git", normalize_links=False)
        copier.copy(str(src_dir), str(dest_dir))
        m.assert_called_with(str(src_dir), git_bin_path="/path/to/git", log_debug=None)
        assert dest_dir.is_dir()
        assert {p.name for p in dest_dir.iterdir()} == {
            "link_dir",
            "trick_link",
            "out_link",
            "out_link_dir",
            "abs_link",
            "dead_link",
        }
        assert_same(src_dir / "link_dir", dest_dir / "link_dir")
        assert_same((src_dir / "trick_link").resolve(), dest_dir / "trick_link")
        assert_same((src_dir / "out_link").resolve(), dest_dir / "out_link")
        assert_same_recursively(
            (src_dir / "out_link_dir").resolve(), dest_dir / "out_link_dir"
        )
        assert_same((src_dir / "abs_link").resolve(), dest_dir / "abs_link")
        assert_same(src_dir / "dead_dir", dest_dir / "dead_dir")

    dest_dir = directory / "dest4"
    with mock.patch(
        "antsibull_fileutils.copier.list_git_files",
        return_value=[
            b"link_dir",
            b"trick_link",
            b"out_link",
            b"out_link_dir",
            b"abs_link",
            b"dead_link",
        ],
    ) as m:
        copier = GitCopier(git_bin_path="/path/to/git")
        copier.copy(str(src_dir), str(dest_dir))
        m.assert_called_with(str(src_dir), git_bin_path="/path/to/git", log_debug=None)
        assert dest_dir.is_dir()
        assert {p.name for p in dest_dir.iterdir()} == {
            "link_dir",
            "trick_link",
            "out_link",
            "out_link_dir",
            "abs_link",
            "dead_link",
        }
        assert_same(src_dir / "link_dir", dest_dir / "link_dir")
        assert_same(
            src_dir / "trick_link",
            dest_dir / "trick_link",
            changed_link=pathlib.Path("empty"),
        )
        assert_same((src_dir / "out_link").resolve(), dest_dir / "out_link")
        assert_same_recursively(
            (src_dir / "out_link_dir").resolve(), dest_dir / "out_link_dir"
        )
        assert_same(
            src_dir / "abs_link",
            dest_dir / "abs_link",
            changed_link=pathlib.Path("empty"),
        )
        assert_same(src_dir / "dead_dir", dest_dir / "dead_dir")

    def src_dst(*appendix):
        s = src_dir
        d = dest_dir
        for a in appendix:
            s = s / a
            d = d / a
        return str(s), str(d)

    dest_dir = directory / "dest5"
    with mock.patch(
        "antsibull_fileutils.copier.list_git_files",
        side_effect=ValueError("nada"),
    ) as m:
        copier = GitCopier(git_bin_path="/path/to/git")
        with pytest.raises(
            CopierError,
            match=f"^Error while listing files not ignored by Git in {re.escape(str(src_dir))}: nada$",
        ) as exc:
            copier.copy(str(src_dir), str(dest_dir))
        m.assert_called_with(str(src_dir), git_bin_path="/path/to/git", log_debug=None)

    dest_dir = directory / "dest6"
    with mock.patch(
        "antsibull_fileutils.copier.list_git_files",
        return_value=[
            b"out_dead_link",
        ],
    ) as m:
        copier = GitCopier(git_bin_path="/path/to/git")
        with pytest.raises(FileNotFoundError) as exc:
            copier.copy(str(src_dir), str(dest_dir))
        m.assert_called_with(str(src_dir), git_bin_path="/path/to/git", log_debug=None)

    dest_dir = directory / "dest7"
    with mock.patch(
        "antsibull_fileutils.copier.list_git_files",
        return_value=[b"link", b"foobar", b"dir/binary_file", b"dir/another_file"],
    ) as m:
        kwargs, debug, info = collect_log(with_info=False)
        copier = GitCopier(git_bin_path="/path/to/git", **kwargs)
        copier.copy(src_dir, dest_dir, exclude_root=["does-not-exist", "link", "dir"])
        m.assert_called_with(src_dir, git_bin_path="/path/to/git", **kwargs)

        assert debug == [
            ("Identifying files not ignored by Git in {!r}", (src_dir,)),
            ("Copying {} file(s) from {!r} to {!r}", (4, src_dir, dest_dir)),
        ]
        assert dest_dir.is_dir()
        assert {p.name for p in dest_dir.iterdir()} == set()


def test_collection_copier(tmp_path_factory):
    src_dir: pathlib.Path = tmp_path_factory.mktemp("collection-copier")

    copier = mock.MagicMock()

    with CollectionCopier(
        source_directory=src_dir, namespace="foo", name="bar", copier=copier
    ) as (root_dir, collection_dir):
        assert (
            os.path.join(root_dir, "collections", "ansible_collections", "foo", "bar")
            == collection_dir
        )
        assert os.path.exists(root_dir)
        assert os.path.exists(
            os.path.join(root_dir, "collections", "ansible_collections", "foo")
        )
        assert not os.path.exists(collection_dir)  # our mock doesn't create it

    assert not os.path.exists(root_dir)
    copier.copy.assert_called_with(src_dir, collection_dir)


def test_collection_copier_fail():
    copier = mock.MagicMock()
    copier.copy = mock.MagicMock(side_effect=CopierError("boo"))

    kwargs, debug, info = collect_log(with_info=False)

    cc = CollectionCopier(
        source_directory="/foo", namespace="foo", name="bar", copier=copier, **kwargs
    )
    assert os.path.exists(cc.dir)

    with pytest.raises(CopierError, match="^boo$"):
        with cc as (root_dir, collection_dir):
            assert False

    copier.copy.assert_called_with(
        "/foo", os.path.join(cc.dir, "collections", "ansible_collections", "foo", "bar")
    )
    assert not os.path.exists(cc.dir)
    assert debug == [
        (
            "Temporary collection directory: {!r}",
            (os.path.join(cc.dir, "collections", "ansible_collections", "foo", "bar"),),
        )
    ]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Ansible-friendly temporary directories.
"""

from __future__ import annotations

import functools
import os
import shutil
import tempfile
import types
import typing as t
from pathlib import Path


def _get_tempdir_proposals() -> t.Generator[Path]:
    yield Path(tempfile.gettempdir())
    for env_var in ("ANTSIBULL_FILEUTILS_TMPDIR",):
        if os.environ.get(env_var):
            yield Path(os.environ[env_var])
    # Try OS specific locations (we stick to *nix systems here):
    yield Path("/tmp")
    yield Path("/var/tmp")
    yield Path("/usr/tmp")
    # Last resort:
    yield Path.cwd()


def is_acceptable_tempdir(directory: Path) -> bool:
    """
    Ensure that the given directory is not inside an ansible_collections tree.
    """
    if directory.name == "ansible_collections":
        return False
    for parent in directory.parents:
        if parent.name == "ansible_collections":
            return False
    return True


def find_tempdir(is_acceptable: t.Callable[[Path], bool]) -> Path:
    """
    Find a temporary directory that is acceptable according to the provided predicate.

    Raises a ``ValueError`` if no appropriate directory can be found.
    """
    for path in _get_tempdir_proposals():
        if not path.is_dir():
            continue
        directory = path.absolute()
        if is_acceptable(directory):
            return directory
    candidates = ", ".join([str(path) for path in _get_tempdir_proposals()])
    raise ValueError(
        f"Cannot find acceptable temporary directory! Candidates: {candidates}"
    )


@functools.cache
def _get_tempdir() -> Path:
    return find_tempdir(is_acceptable_tempdir)


def ansible_mkdtemp(*, suffix: str | None = None, prefix: str | None = None) -> Path:
    """
    Create a temporary directory that is guaranteed
    not to be inside an ``ansible_collections`` tree.

    The ``suffix`` and ``prefix`` directories behave as for ``tempfile.mkdtemp()``.
    Will raise ``ValueError`` if no temporary directory can be found.
    This can only happen if places like ``/tmp``, ``/var/tmp``, and ``/usr/tmp``
    all do not exist.
    """
    result = Path(tempfile.mkdtemp(suffix=suffix, prefix=prefix, dir=_get_tempdir()))
    if not is_acceptable_tempdir(result):
        # Clean up before erroring out
        try:
            result.unlink(missing_ok=True)
        except:  # noqa: E722 # pylint: disable=bare-except
            pass
        raise ValueError(f"Internal error: got invalid temp directory {result}")
    return result


class TemporaryDirectoryHelper:
    def __init__(
        self,
        *,
        directory: Path,
        ignore_cleanup_errors: bool = False,
        delete: bool = True,
    ) -> None:
        """
        Provide a temporary directory as a context manager.

        Note that opposed to ``tempfile.TemporaryDirectory``, no finalizer is registered
        that will clean up the temporary directory if garbage collector cleans up the
        class before ``__exit__()`` was called.
        """
        self._ignore_cleanup_errors = ignore_cleanup_errors
        self._delete = delete
        self._directory = directory

    @property
    def name(self) -> Path:
        """
        Return the temporary directory.
        """
        return self._directory

    def cleanup(self) -> None:
        """
        Remove temporary directory.
        """
        if not self._directory.exists():
            return
        shutil.rmtree(self._directory, ignore_errors=self._ignore_cleanup_errors)

    def __enter__(self) -> Path:
        return self._directory

    def __exit__(
        self,
        exc_type: t.Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> t.Literal[False]:
        if self._delete:
            self.cleanup()
        return False

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.name!r}>"


class AnsibleTemporaryDirectory(TemporaryDirectoryHelper):
    def __init__(
        self,
        *,
        suffix: str | None = None,
        prefix: str | None = None,
        ignore_cleanup_errors: bool = False,
        delete: bool = True,
    ) -> None:
        """
        Create a temporary directory and provide it as a context manager.

        Note that opposed to ``tempfile.TemporaryDirectory``, no finalizer is registered
        that will clean up the temporary directory if garbage collector cleans up the
        class before ``__exit__()`` was called.
        """
        super().__init__(
            directory=ansible_mkdtemp(suffix=suffix, prefix=prefix),
            ignore_cleanup_errors=ignore_cleanup_errors,
            delete=delete,
        )


__all__ = (
    "ansible_mkdtemp",
    "find_tempdir",
    "is_acceptable_tempdir",
    "AnsibleTemporaryDirectory",
    "TemporaryDirectoryHelper",
)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Test tempfile module.
"""

from __future__ import annotations

import contextlib
import os
import pathlib
import re
from pathlib import Path
from unittest import mock

import pytest

from antsibull_fileutils.tempfile import (
    TemporaryDirectoryHelper,
    find_tempdir,
    is_acceptable_tempdir,
)


@pytest.mark.parametrize(
    "directory, expected",
    [
        ("/", True),
        ("/ansible_collections", False),
        ("/ansible_collections/foo", False),
        ("/ansible_collections/foo/bar", False),
        ("/foo/ansible_collections", False),
        ("/foo/bar/ansible_collections", False),
        ("/foo/ansible_collections/bar", False),
        ("/foo/bar/ansible_collections/baz/bam", False),
        ("/foo/bar/baz/bam", True),
    ],
)
def test_is_acceptable_tempdir(directory: str, expected: bool) -> None:
    assert is_acceptable_tempdir(Path(directory)) == expected


def set_env_var_value(env_var: str, value: str | None) -> None:
    if value is None:
        os.environ.pop(env_var, None)
    else:
        os.environ[env_var] = value


@contextlib.contextmanager
def set_env_var(env_var: str, value: str | None):
    current_value = os.environ.get(env_var)
    try:
        set_env_var_value(env_var, value)
        yield
    finally:
        set_env_var_value(env_var, current_value)


def test_find_tempdir(tmp_path: Path) -> None:
    random_dirname = "fooThaequie4QuaP3tie"

    def is_acceptable(path: Path) -> bool:
        print(path)
        if path.name == random_dirname:
            return True
        for parent in path.parents:
            if parent.name == random_dirname:
                return True
        return False

    random_dir = tmp_path / random_dirname / "bar"
    random_dir.mkdir(parents=True)
    with set_env_var("ANTSIBULL_FILEUTILS_TMPDIR", str(random_dir)):
        print(os.environ)
        assert find_tempdir(is_acceptable) == random_dir

    with set_env_var("ANTSIBULL_FILEUTILS_TMPDIR", None):
        with pytest.raises(ValueError):
            find_tempdir(is_acceptable)

    with pytest.raises(ValueError):
        find_tempdir(lambda path: False)


def test_TemporaryDirectoryHelper(tmp_path_factory) -> None:
    temp: Path = tmp_path_factory.mktemp("temp")

    assert temp.is_dir()
    tdh = TemporaryDirectoryHelper(directory=temp)
    assert tdh.name == temp
    with tdh:
        assert temp.is_dir()
        assert tdh.name == temp
    assert not temp.exists()
    assert not temp.is_dir()
    assert tdh.name == temp

    temp.mkdir()
    assert temp.is_dir()
    with TemporaryDirectoryHelper(directory=temp):
        assert temp.is_dir()
        temp.rmdir()
    assert not temp.exists()
    assert not temp.is_dir()

    temp.mkdir()
    assert temp.is_dir()
    with TemporaryDirectoryHelper(directory=temp, delete=False):
        assert temp.is_dir()
    assert temp.is_dir()

    assert temp.is_dir()
    with TemporaryDirectoryHelper(directory=temp, delete=False):
        assert temp.is_dir()
        temp.rmdir()
    assert not temp.exists()
    assert not temp.is_dir()



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
Git functions.
"""

from __future__ import annotations

import subprocess
import typing as t

if t.TYPE_CHECKING:
    from _typeshed import StrPath


def detect_vcs(
    path: StrPath,
    *,
    git_bin_path: StrPath = "git",
    log_debug: t.Callable[[str], None] | None = None,
    log_info: t.Callable[[str], None] | None = None,
) -> t.Literal["none", "git"]:
    """
    Try to detect whether the given ``path`` is part of a VCS repository.

    NOTE: The return type might be extended in the future. To be on the safe
          side, test for the types you support, and use a fallback for unknown
          values (treat them like ``"none"``).
    """

    def do_log_debug(msg: str, *args: t.Any) -> None:
        if log_debug:
            log_debug(msg, *args)

    def do_log_info(msg: str, *args: t.Any) -> None:
        if log_info:
            log_info(msg, *args)

    do_log_debug("Trying to determine whether {!r} is a Git repository", path)
    try:
        result = subprocess.check_output(
            [str(git_bin_path), "-C", path, "rev-parse", "--is-inside-work-tree"],
            text=True,
            encoding="utf-8",
            stderr=subprocess.PIPE,
        ).strip()
        do_log_debug("Git output: {}", result)
        if result == "true":
            do_log_info("Identified {!r} as a Git repository", path)
            return "git"
    except subprocess.CalledProcessError as exc:
        # This is likely not inside a work tree
        do_log_debug("Git failed: {}", exc)
    except FileNotFoundError as exc:
        # Cannot find git executable
        do_log_debug("Cannot find git: {}", exc)

    # Fallback: no VCS detected
    do_log_debug("Cannot identify VCS")
    return "none"


def list_git_files(
    directory: StrPath,
    *,
    git_bin_path: StrPath = "git",
    log_debug: t.Callable[[str], None] | None = None,
) -> list[bytes]:
    """
    List all files not ignored by git in a directory and subdirectories.

    Raises ``ValueError`` in case of errors.
    """

    def do_log_debug(msg: str, *args) -> None:
        if log_debug:
            log_debug(msg, *args)

    do_log_debug("Identifying files not ignored by Git in {!r}", directory)
    try:
        result = subprocess.check_output(
            [
                str(git_bin_path),
                "ls-files",
                "-z",
                "--cached",
                "--others",
                "--exclude-standard",
                "--deduplicate",
            ],
            cwd=directory,
            stderr=subprocess.PIPE,
        ).strip(b"\x00")
        if result == b"":
            return []
        return result.split(b"\x00")
    except subprocess.CalledProcessError as exc:
        raise ValueError("Error while running git") from exc
    except FileNotFoundError as exc:
        raise ValueError("Cannot find git executable") from exc



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
Utilities
"""

from __future__ import annotations


def collect_log(with_debug: bool = True, with_info: bool = True):
    debug = []
    info = []

    def log_debug(msg: str, *args) -> None:
        debug.append((msg, args))

    def log_info(msg: str, *args) -> None:
        info.append((msg, args))

    kwargs = {}
    if with_debug:
        kwargs["log_debug"] = log_debug
    if with_info:
        kwargs["log_info"] = log_info
    return (
        kwargs,
        debug,
        info,
    )
